# pylint: disable=E1101

# imports - compatibility imports
from pipupgrade         import _compat
from pipupgrade._compat import zip

# imports - standard imports
import sys
import inspect

def merge_dict(*args):
    merged = dict()

    for arg in args:
        copy = arg.copy()
        merged.update(copy)

    return merged
    
def list_filter(v, filter_):
    filtered = filter(filter_, v)
    filtered = list(filtered)

    return filtered

def dict_from_list(keys, values):
    return dict(zip(keys, values))

def get_function_arguments(fn):
    # https://stackoverflow.com/a/2677263
    params  = dict()
    success = False

    if _compat.PY2:
        argspec_getter = inspect.getargspec
        success        = True
    if _compat.PYTHON_VERSION >= (3,0) and (3,4) <= _compat.PYTHON_VERSION:
        argspec_getter = inspect.getfullargspec
        success        = True

    if success:
        argspec   = argspec_getter(fn)
        params    = dict_from_list(argspec.args, argspec.defaults or [])

    if _compat.PYTHON_VERSION >= (3,5):
        signature  = inspect.signature(fn)
        parameters = signature.parameters

        params     = { k: v.default for k, v in _compat.iteritems(parameters) }

        success    = True

    if not success:
        raise ValueError("Unknown Python Version {} for fetching functional arguments.".format(sys.version))

    return params


# imports - compatibility imports
from pipupgrade._compat import iteritems

# imports - standard imports
import sys, os, os.path as osp
import re
from subprocess import call, list2cmdline

# imports - module imports
from pipupgrade.commands.util import cli_format
from pipupgrade.table      	  import Table
from pipupgrade.util.types    import list_filter
from pipupgrade.util.string   import pluralize
from pipupgrade.util.system   import read, write
from pipupgrade 		      import _pip, request as req, cli, semver

_SEMVER_COLOR_MAP = dict(
	major = cli.RED,
	minor = cli.YELLOW,
	patch = cli.GREEN
)

def _cli_format_semver(version, type_):
	def _format(x):
		return cli_format(x, cli.YELLOW)
	
	try:
		semver.parse(version)
		
		if type_ == "major":
			version    = _format(version)
		if type_ == "minor":
			index      = version.find(".", 1) + 1
			head, tail = version[:index], version[index:]
			version    = "".join([head, _format(tail)])
		if type_ == "patch":
			index      = version.find(".", 2) + 1
			head, tail = version[:index], version[index:]
			version    = "".join([head, _format(tail)])
	except ValueError:
		pass

	return version

def _get_pypi_info(name, raise_err = True):
	url  = "https://pypi.org/pypi/{}/json".format(name)
	res  = req.get(url)

	info = None

	if res.ok:
		data = res.json()
		info = data["info"]
	else:
		if raise_err:
			res.raise_for_status()

	return info

class PackageInfo:
	def __init__(self, package):
		if   isinstance(package, (_pip.Distribution, _pip.DistInfoDistribution, _pip.EggInfoDistribution)):
			self.name            = package.project_name
			self.current_version = package.version
		elif isinstance(package, _pip.InstallRequirement):
			self.name            = package.name
			self.current_version = package.installed_version

		_pypi_info = _get_pypi_info(self.name, raise_err = False) or { }

		self.latest_version = _pypi_info.get("version")
		self.home_page      = _pypi_info.get("home_page")

def _update_requirements(path, package):
	path 	= osp.realpath(path)
	
	content = read(path)
		
	try:
		pattern = r"{package}(=={version})*".format(
			package = re.escape(package.name),
			version = re.escape(package.current_version)
		)
		lines   = content.splitlines()
		
		with open(path, "w") as f:
			for line in lines:
				if re.search(pattern, line, flags = re.IGNORECASE):
					line = line.replace(
						"==%s" % package.current_version,
						"==%s" % package.latest_version
					)
					
					f.write(line)
	except Exception:
		write(path, content)

@cli.command
def command(requirements = [ ], latest = False, user = False, check = False, yes = False, no_color = True, verbose = False):
	cli.echo(cli_format("Checking...", cli.YELLOW))
	
	registry = dict()
	
	if requirements:
		for requirement in requirements:
			path = osp.realpath(requirement)

			if not osp.exists(path):
				cli.echo(cli_format("{} not found.".format(path), cli.RED))
				sys.exit(os.EX_NOINPUT)
			else:
				registry[path] = _pip.parse_requirements(requirement, session = "hack")
	else:
		registry["__INSTALLED__"] = _pip.get_installed_distributions()

	for source, packages in iteritems(registry):
		table = Table(header = ["Name", "Current Version", "Latest Version", "Home Page"])
		dinfo = [ ] # Information DataFrame

		for package in packages:
			package = PackageInfo(package)

			if package.latest_version and package.current_version != package.latest_version:
				diff_type = None

				try:
					diff_type = semver.difference(package.current_version, package.latest_version)
				except ValueError:
					pass

				table.insert([
					cli_format(package.name, _SEMVER_COLOR_MAP.get(diff_type, cli.CLEAR)),
					package.current_version,
					_cli_format_semver(package.latest_version, diff_type),
					cli_format(package.home_page, cli.CYAN)
				])

				package.source    = source
				package.diff_type = diff_type

				dinfo.append(package)

		stitle = "Installed Distributions" if source == "__INSTALLED__" else source

		if not table.empty:
			string = table.render()
		
			cli.echo("\nSource: %s\n" % stitle)
			cli.echo(string)
			cli.echo()

			if not check:
				npackages = len(dinfo) - (len([p for p in dinfo if p.diff_type == "major"]) if not latest else 0)
				spackages = pluralize("package", npackages) # Packages "string"

				query 	  = "Do you wish to update %s %s?" % (npackages, spackages)
				if npackages and (yes or cli.confirm(query)):
					for i, package in enumerate(dinfo):
						if package.diff_type != "major" or latest:
							cli.echo(cli_format(
								"Updating %s of %s %s: %s" % (
									i + 1,
									npackages,
									spackages,
									cli_format(package.name, cli.GREEN)
								)
							, cli.BOLD))

							params  = list_filter([
								"pip",
								"install",
								"--user"  if user 		 else None,
								"--quiet" if not verbose else None,
								"--no-cache",
								"--upgrade",
								package.name,
							], filter_ = bool)
							command = list2cmdline(params)
							
							call(command, shell = True)

							if package.source != "__INSTALLED__":
								_update_requirements(package.source, package)
		else:
			cli.echo("%s upto date." % cli_format(stitle, cli.CYAN))


# imports - compatibility imports
from pipupgrade._compat import cmp

# imports - standard imports
import re

_REGEX_SEMVER = re.compile(r"""
    ^
    (?P<major>(?:0|[1-9][0-9]*))
    \.
    (?P<minor>(?:0|[1-9][0-9]*))
    \.
    (?P<patch>(?:0|[1-9][0-9]*))
    (\-(?P<prerelease>
        (?:0|[1-9A-Za-z-][0-9A-Za-z-]*)
        (\.(?:0|[1-9A-Za-z-][0-9A-Za-z-]*))*
    ))?
    (\+(?P<build>
        [0-9A-Za-z-]+
        (\.[0-9A-Za-z-]+)*
    ))?
    $
    """, re.VERBOSE)

def parse(version):
    match   = _REGEX_SEMVER.match(version)
    if not match:
        raise ValueError("{} is not a valid Semantic Version string".format(version))

    version = match.groupdict()

    version["major"] = int(version["major"])
    version["minor"] = int(version["minor"])
    version["patch"] = int(version["patch"])

    return version

def difference(a, b):
    va = parse(a)
    vb = parse(b)

    if a != b:
        for key in ["major", "minor", "patch"]:
            if cmp(va.get(key), vb.get(key)):
                return key
        
        raise NotImplementedError("Unknown difference between {} and {}".format(a, b))
    else:
        return None


# imports - compatibility imports
from pipupgrade._compat import HTTPError

# imports - standard imports
import json

_HTTP_RESPONSE_TYPE = {
    404: { "name": "Not Found" }
}

class Response:
    def __init__(self):
        self.status_code = None
        self.content     = None
        self.url         = None
        
    @property
    def ok(self):
        status_code = self.status_code
        
        if status_code < 400:
            return True
        else:
            return False
        
    def json(self):
        content = self.content
        string  = content.decode()

        json_   = json.loads(string)

        return json_

    def __repr__(self):
        string = "<Response [{}]>".format(self.status_code)
        return string

    def raise_for_status(self):
        if not self.ok:
            code  = self.status_code

            name  = _HTTP_RESPONSE_TYPE[code]["name"]
            error = "{} for URL {}".format(name, self.url)

            raise HTTPError(self.url, code, error, hdrs = { }, fp = None)


# imports - standard imports
import argparse

# imports - module imports
from pipupgrade.__attr__ import (
    __version__,
    __description__,
    __command__
)

def get_parser():
    parser = argparse.ArgumentParser(
        prog        = __command__,
        description = __description__,
        add_help    = False
    )
    parser.add_argument("-y", "--yes",
        action  = "store_true",
        help    = "Confirm for all dialogs"
    )
    parser.add_argument("-c", "--check",
        action  = "store_true",
        help    = "Check for outdated packages"
    )
    parser.add_argument("-l", "--latest",
        action  = "store_true",
        help    = "Update all packages to latest"
    )
    parser.add_argument("-r", "--requirements",
        action  = "append",
        help    = "Path to requirements.txt file"
    )
    parser.add_argument("-u", "--user",
        action  = "store_true",
        help    = "Install to the Python user install directory for environment \
                    variables and user configuration."
    )
    parser.add_argument("--no-color",
        action  = "store_true",
        help    = "Avoid colored output"
    )
    parser.add_argument("-V", "--verbose",
        action  = "store_true",
        help    = "Display verbose output"
    )

    parser.add_argument("-v", "--version",
        action  = "version",
        version = __version__
    )
    parser.add_argument("-h", "--help",
        action  = "help",
        default = argparse.SUPPRESS,
        help    = "Show this help message and exit"
    )

    return parser

def get_args(args = None, known = True, as_dict = True):
    parser  = get_parser()

    if known:
        args, _ = parser.parse_known_args(args)
    else:
        args    = parser.parse_args(args)

    if as_dict:
        args = args.__dict__
        
    return args


# imports - compatibility imports
from __future__ import print_function
from pipupgrade._compat import input

# imports - standard imports
import inspect

# imports - module imports
from pipupgrade.cli.parser import get_args
from pipupgrade.util.types import merge_dict, get_function_arguments

_ACCEPTABLE_YES_INPUTS = ("", "y", "Y")

_ANSI_FORMAT = "\033[{}m"
_format_ansi = lambda x: _ANSI_FORMAT.format(x)

BOLD      = _format_ansi("0;1")
UNDERLINE = _format_ansi("0;4")
RED       = _format_ansi("0;91")
GREEN     = _format_ansi("0;92")
YELLOW    = _format_ansi("0;93")
CYAN      = _format_ansi("0;96")
CLEAR     = _format_ansi("0")

def confirm(query):
    query  = "{} [Y/n]: ".format(query)
    output = input(query)

    return output in _ACCEPTABLE_YES_INPUTS

def format(string, type_):
    string = "{}{}{}".format(type_, string, CLEAR)
    return string

def echo(string = "", nl = True):
    print(string, end = "\n" if nl else "")

def command(fn):
    args    = get_args()
    
    params  = get_function_arguments(fn)

    params  = merge_dict(params, args)
    
    def wrapper(*args, **kwargs):
        return fn(**params)

    return wrapper


# imports - standard imports
import os.path as osp

def read(fname):
    with open(fname) as f:
        data = f.read()
    return data

def write(fname, data = None, force = False):
    if not osp.exists(fname) or force:
        with open(fname, "w") as f:
            if data:
                f.write(data)


# imports - compatibility imports
from pipupgrade._compat     import zip_longest

# imports - module imports
from pipupgrade.util.string import strip_ansi

def _sanitize_string(string):
    """
    Helper for tabulate to sanitize strings.
    """
    string = strip_ansi(string)
    return string

def tabulate(rows):
    # Shamelessly taken from: https://git.io/fARTL (pip)
    # Also: https://git.io/fARTY (yarn)
    # Yo, WTF is wrong with git.io and its "fart" filled short urls?

    sizes  = [0] * max(len(x) for x in rows)
    for row in rows:
        sizes = [max(s, len(str(_sanitize_string(c if c else "")))) for s, c in zip_longest(sizes, row)]

    result = [ ]
    for row in rows:
        display = " ".join([str(c) + " " * (s - len(_sanitize_string(c if c else ""))) if c is not None else ""
                            for s, c in zip_longest(sizes, row)])
        result.append(display)

    return result, sizes

class Table:
    def __init__(self, header = None):
        self.rows   = [ ]
        self.header = header or [ ]

    @property
    def empty(self):
        _empty = len(self) == 0
        return _empty

    def insert(self, row):
        self.rows.append(row)

    def render(self, header = True):
        string = ""

        _rows  = self.rows[:]
        
        if header and self.header:
            _rows.insert(0, self.header)

        if _rows:
            tabulated, sizes = tabulate(_rows)
            
            # divider
            if header:
                tabulated.insert(1, " ".join(map(lambda x: "-" * x, sizes)))

            string  = "\n".join(tabulated)

        return string
    
    def __len__(self):
        length = len(self.rows)
        return length


# imports - module imports
from pipupgrade.cli.parser import get_args
from pipupgrade import cli

def cli_format(string, type_):
    args = get_args(as_dict = False)

    if not args.no_color:
        string = cli.format(string, type_)

    return string


# imports - standard imports
import pip

PIP9 = int(pip.__version__.split(".")[0]) < 10

if PIP9:
    from pip                 import get_installed_distributions
    from pip.req             import parse_requirements
    from pip.req.req_install import InstallRequirement
else:
    from pip._internal.utils.misc      import get_installed_distributions
    from pip._internal.req             import parse_requirements
    from pip._internal.req.req_install import InstallRequirement

from pip._vendor.pkg_resources import (
    Distribution,
    DistInfoDistribution,
    EggInfoDistribution
)


# imports - standard imports
import re

_REGEX_ANSI_ESCAPE = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")

def strip_ansi(string):
    string = _REGEX_ANSI_ESCAPE.sub("", string)
    return string

def pluralize(string, count = 1):
    # A very shitty pluralizer
    if not string.endswith("s"):
        if count > 1:
            string += "s"
    
    return string


# imports - compatibility imports
from pipupgrade._compat import StringIO, input

# imports - standard imports
import sys
from   contextlib import contextmanager

__STDIN__ = sys.stdin

@contextmanager
def mock_input(args):
    # https://stackoverflow.com/a/36491341
    sys.stdin = args
    yield
    sys.stdin = __STDIN__

def assert_stdout(capfd, output):
    stdout, _ = capfd.readouterr()
    assert output == stdout

def assert_input(capfd, text, output, expected = None, input_ = None, stdout = None):
    if expected == None:
        expected = output
    input_ = input_ or input
    stdout = stdout or text
    
    with mock_input(StringIO(output)):
        assert input_(text) == expected
        assert_stdout(capfd, stdout)


import sys
import os.path as osp
import subprocess

PY2 = sys.version_info.major == 2

try:
    FileNotFoundError
except NameError:
    if PY2:
        FileNotFoundError = OSError

def read(fname):
    with open(fname) as f:
        data = f.read()
    return data

def pardir(fname, level = 1):
    for _ in range(level):
        fname = osp.dirname(fname)
    return fname

def strip(string):
    string = string.lstrip()
    string = string.rstrip()
    return string

def safe_decode(object_, encoding = "utf-8"):
    decoded = object_
    
    try:
        decoded = object_.decode(encoding)
    except AttributeError:
        pass

    return decoded

def sequence_filter(list_, filter_, type_ = list):
    result = type_(filter(filter_, list_))
    return result

def get_revision(path, short = False, raise_err = True):
    """
    Returns the git revision of a repository. Raises error if not a valid git repository.
    """
    revision = None

    try:
        short    = "--short" if short else ""
        output   = subprocess.check_output(sequence_filter(["git", "rev-parse", short, "HEAD"], filter_ = None), cwd = path)
        revision = safe_decode(strip(output))
    except (subprocess.CalledProcessError, FileNotFoundError):
        if raise_err:
            raise

    return revision

path                        = dict()
path["base"]                = pardir(__file__)
path["version"]             = osp.join(path["base"], "VERSION")

__name__                    = "pipupgrade"
__command__                 = __name__
__version__                 = read(path["version"])
__build__                   = get_revision(pardir(path["base"], 2), short = True, raise_err = False)
__url__                     = "https://github.com/achillesrasquinha/pipupgrade"
__author__                  = "Achilles Rasquinha"
__email__                   = "achillesrasquinha@gmail.com"
__description__             = "UPGRADE ALL THE PIP PACKAGES!"
__keywords__                = ["pip", "update", "upgrade", "cli", "command"]
__license__                 = "MIT"


# pylint: disable=E0602

import sys
import os.path as osp
import glob

from   setuptools import setup, find_packages

import pip

try:
    from pip._internal.req import parse_requirements # pip 10
except ImportError:
    from pip.req           import parse_requirements # pip 9

# globals
PACKAGE     = "pipupgrade"
SRCDIR      = "src"

def isdef(var):
    return var in globals()

def read(path):
    content = None
    
    with open(path) as f:
        content = f.read()

    return content

def get_package_info():
    attr = osp.join(SRCDIR, PACKAGE, "__attr__.py")
    info = dict(__file__ = attr) # HACK
    
    with open(attr) as f:
        content = f.read()
        exec(content, info)

    return info

def get_dependencies(type_ = None):
    path         = osp.realpath("requirements.txt")
    requirements = [str(ir.req) for ir in parse_requirements(path, session = "hack")]

    if type_ == "development":
        path         = osp.realpath("requirements-dev.txt")
        requirements = [
            str(ir.req) for ir in parse_requirements(path, session = "hack")
                if str(ir.req) not in requirements
        ]
    
    return requirements

PKGINFO    = get_package_info()

setup(
    name                 = PKGINFO["__name__"],
    version              = PKGINFO["__version__"],
    url                  = PKGINFO["__url__"],
    author               = PKGINFO["__author__"],
    author_email         = PKGINFO["__email__"],
    description          = PKGINFO["__description__"],
    long_description     = read("README.md"),
    license              = PKGINFO["__license__"],
    keywords             = " ".join(PKGINFO["__keywords__"]),
    packages             = find_packages(where = SRCDIR, exclude = ["tests"]),
    package_dir          = { "": SRCDIR },
    entry_points         = {
        "console_scripts": [
            "%s = %s.__main__:main" % (
                PKGINFO["__command__"] if hasattr(PKGINFO, "__command__") else PKGINFO["__name__"],
                PACKAGE
            )
        ]
    },
    
    install_requires     = get_dependencies(type_ = "production"),
    extras_require       = dict(
        dev = get_dependencies(type_ = "development")
    ),
    include_package_data = True,
    classifiers          = (
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy"
    )
)



# imports - standard imports
import sys

PYTHON_VERSION = sys.version_info

def _is_python_version(*args, **kwargs):
    major  = kwargs.get("major", None)
    minor  = kwargs.get("minor", None)
    patch  = kwargs.get("patch", None)

    result = True

    if major:
        result = result and major == PYTHON_VERSION.major
    if minor:
        result = result and minor == PYTHON_VERSION.minor
    if patch:
        result = result and patch == PYTHON_VERSION.micro
        
    return result

PY2 = _is_python_version(major = 2)

def cmp(a, b):
    return ((a > b) - (a < b))

def iteritems(dict_, **kwargs):
    if PY2:
        iterator = dict_.iteritems()
    else:
        iterator = iter(dict_.items(), **kwargs)
    return iterator

def iterkeys(dict_, **kwargs):
    if PY2:
        iterator = dict_.iterkeys()
    else:
        iterator = iter(dict_.keys(), **kwargs)
    return iterator

if PY2:
    import __builtin__ as builtins

    # moves
    from urllib2 import urlopen
    from urllib2 import HTTPError

    from __builtin__ import raw_input as input

    from StringIO import StringIO

    from itertools import izip         as zip
    from itertools import izip_longest as zip_longest
else:
    import builtins

    # moves
    from urllib.request import urlopen
    from urllib.error   import HTTPError

    from builtins import input

    from io import StringIO

    from itertools import zip_longest

    zip = zip


# imports - compatibility imports
from pipupgrade._compat import urlopen, HTTPError

# imports - standard imports
import json

# imports - module imports
from pipupgrade.request.response import Response

def get(url):
    response = Response()

    try:
        http_response    = urlopen(url)
        status_code      = http_response.getcode()

        response.content = http_response.read()

        http_response.close()
    except HTTPError as e:
        status_code      = e.getcode()

    response.url         = url
    response.status_code = status_code

    return response